/*
Program Description:

This program reads in a data file and generates a trace of dictionary 
operations that can be used by the DIMACS Dictionary driver. 
The input file parameters are as follows:
    ini X            number of initial inserts 
    rep Y            number of repetitions
    ins              include inserts in repetitions
    suc              include successful lookups in repetitions
    uns              include unsuccessful lookups
    del              include deletes
    sed R            optional random seed
    key K            integer keyparts between 0..K-1
    end N            end of parameters, keys start below

Parameters must be followed by a list containing _UNIQUE_ keys (presumably 
strings), one per line, each having maximum length N. 

The program generates X inserts using the keylist, followed by Y
repetitions of whichever operations in [ins suc uns del] are specified 
by user.  The generator uses key from the file for its inserts and 
unsuccessful lookups, so there must be `enough' to cover those operations 
(either X, X+Y, or X+2Y depending on which are included in reps).  

Keys for successful lookups and deletes are obtained by finding a random
item in the dictionary. 

Written by Marc Brooks with some modifications by C. McGeoch  8/96

*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include "dict.h"

cmdtype cmdtable[CMD_SIZE+1];    /* table of user commands               */
int namer;                       /* for unique key timestamps */ 

/*----table_init---------------------------------------------------------*/
void table_init ( ) {
/*
Function Description:
This is simple an initialization function. It is used in the parse() 
function.
*/
  strcpy(cmdtable[0], "com");
  strcpy(cmdtable[1], "ins");   /* whether to use insert in loop */
  strcpy(cmdtable[2], "suc");   /* whether to do successful lookups */
  strcpy(cmdtable[3], "uns");   /* whether to do unsuccess  lookups */
  strcpy(cmdtable[4], "del");   /* whether to do deletes  */ 
  strcpy(cmdtable[5], "rep");    /* how many reps in loop */
  strcpy(cmdtable[6], "ini");    /* how many initial inserts */ 
  strcpy(cmdtable[7], "sed");    /* optional seed input */ 
  strcpy(cmdtable[8], "key");    /* optional max key value */ 
  strcpy(cmdtable[9], "end");    /* end of parameters*/ 
} /* end of table_init */

/*-----param_init--------------------------------------------------------*/ 
void param_init(Pparameters param) {

  param->num_ops = default_num_ops; 
  param->init_ins = default_init_ins;
  param->ins = default_ins;
  param->suc_lookup = default_suc_lookup;
  param->unsuc_lookup = default_unsuc_lookup;
  param->del = default_del;
  param->myseed = (int) time (0);
  param->keysize = DEF_KEYSIZE; 

} /* end of param_init */

/*------parse-----------------------------------------------------------*/
int parse( char* cmd ) {
/*
Function Description:
This function takes a command and return a integer index. 
*/

  int i;
  int go = 1;

  for (i = CMD_SIZE ; go != 0; i--) {
    if (i <0 ) {
         printf("Ignored Unknown Command %s. \n", cmd); return(0);}
    go = strncmp(cmdtable[i], cmd, 3 ); 
  }
  return i+1;
}/*lookup*/

/*---read_paramters---------------------------------------------------*/ 
void read_parameters ( Pparameters param) {
/*

Function Description:
This function reads through the input and writes the user supplied 
parameters to the param.

*/

  cmdtype cmd;
  char tmp[80];
  int n; 

  while ( scanf("%s", cmd) != EOF ) {
    fgets ( tmp, sizeof(tmp), stdin) ; 
    switch ( parse( cmd ) ) {
    case 0 : { /* comment */
	       break;
	     } /* end of case 0 */
    case 1 : { /* use insert */ 
               param->ins = 1;
	       break;
	     } /* end of case 1 */
    case 2 : { /* do successful */
	       param->suc_lookup = 1;
	       break;
	     } /* end of case 2 */
    case 3 : { /* do unsuccessful */
               param->unsuc_lookup = 1;
	       break;
	     } /* end of case 3 */
    case 4 : { /* do deletes */
               param->del = 1;
	       break;
	     } /* end of case 4 */
    case 5 : {/*number repetitions */
              sscanf (tmp, "%d", &n);
              param->num_ops = n; 
              break;
	    } /* end of case 5*/
    case 6 : {/*number of initial inserts*/
              sscanf (tmp, "%d", &n );
              param->init_ins = n;
              break;
	    }/* end of case 6*/
    case 7 : {/* user-supplied seed */
             sscanf (tmp, "%d", &n);
             param -> myseed = n;
             break;
           } /*end of case 7 */
    case 8 : {/* max intpart of key */
             /* not used */ 
             break;
           } /* end of case 8 */
    case 9: { /* end of parameters */
             sscanf (tmp, "%d", &n);
             param->keysize = n; 
             if (n > DEF_KEYSIZE) { 
                printf("Sorry, keysize is too big. Please recompile\n");
                printf("with bigger DEF_KEYSIZE (in dict.h) \n");
                exit(1); }
             return;
             break;
	   } /* end of case 9*/ 
    } /* end of switch statement */
  } /* end of while loop */
  printf ("ERROR: No end line encountered in file\n");
  exit(1);
} /* end of read_parameters */

/*---write_param--------------------------------------------------*/ 
void write_param( Pparameters param ) {

  printf(  "com Dictionary operations with user-provided keys. \n");
  printf(  "com Initial Inserts = %d, repetitions = %d\n",
            param->init_ins, param->num_ops); 
  printf(  "com Insert? %d\n", param->ins);
  printf(  "com Successful Lookup? %d\n", param->suc_lookup);
  printf(  "com Unsuccessful Lookup? %d\n", param->unsuc_lookup);
  printf(  "com Delete? %d\n", param->del); 
  printf(  "com Seed = %d\n", param->myseed); 

} /* end of print_param */

/*---init_dict------------------------------------------------------*/ 
void init_dict ( Pkey_store key_holder, Pparameters param ) {
/* initial inserts in dictionary */

  int i;
  for ( i = 0; i <= key_holder->last_key; i++ ) {
       printf("ins %s\n", key_holder->keys[i] );
      fflush(stdout); 
  } /* end of for loop */

} /* end of init_dict */

/*-----generate_key---------------------------------------------------*/ 
void generate_key( keytype thekey ) {
/* Gets the next key from the input file */ 

    if( gets(thekey) == NULL) {
         printf("Ran out of keys. Stop \n");
         exit(1);
    } else {
      return; 
    }
} 

/*-----generate_init_keys---------------------------------------------*/ 
void generate_init_keys ( Pkey_store key_holder, Pparameters param ) {

  int i;
  keytype thekey; 

  for ( i = 0; i < param->init_ins ; i++ ) {
      generate_key( thekey ) ; 
      strcpy(key_holder->keys[i], thekey); 
  } /* end of for loop */
  key_holder->last_key = param->init_ins -1;

} /* end of generate_init_keys */

/*---write_ins--------------------------------------------------------*/ 
void write_ins ( Pkey_store key_holder, Pparameters param ) {

keytype thekey;

  generate_key( thekey) ; 
  strcpy(key_holder->keys[(key_holder->last_key+1)], thekey); 
  key_holder->last_key++;
  printf ( "ins %s\n", key_holder->keys[key_holder->last_key] );
 
} /* end of write_ins */

/*---write_suc_lookup------------------------------------------------*/
void write_suc_lookup ( Pkey_store key_holder ) {
/* Chooses a random key from list of dictionary contents.            */

  printf ( "lkp %s\n", 
	  (key_holder->keys[(int)((key_holder->last_key + 1) * 
					  drand48())]) );

} /* end write_suc_lookup */

/*---write_unsuc_lookup---------------------------------------------*/ 

void write_unsuc_lookup ( Pkey_store key_holder, Pparameters param ) {

int tmp;
keytype thekey; 
    /* get a valid one and make an impossible timestamp */ 
    
    generate_key(thekey);     
    printf("lkp %s \n", thekey); 

} /* end of write_unsuc_lookup */

/*---key_shuffle------------------------------------------------------*/ 
void key_shuffle ( int doomed_key, int last_key, Pkey_store key_holder ) {

  strcpy(key_holder->keys[doomed_key], key_holder->keys[last_key]); 
  key_holder->last_key -= 1;

} /* end of swap */

/*---write_del---------------------------------------------------------*/
void write_del ( Pkey_store key_holder ) {

  int doomed_key;
  
  doomed_key = (key_holder->last_key+1) * drand48();
  printf( "del %s\n", key_holder->keys[doomed_key] );
  key_shuffle( doomed_key, key_holder->last_key, key_holder );

} /* end of write_del */

/*---write_outfile---------------------------------------------------*/
void write_outfile ( Pparameters param ) {

  Pkey_store key_holder;
  int inserted[param->num_ops];
  int i;

  key_holder = (Pkey_store)malloc(sizeof(key_store));

  generate_init_keys( key_holder , param);

  write_param( param );

  init_dict( key_holder, param );

  for ( i = 0; i < param->num_ops; i++ ) {
    if ( param->ins )        write_ins ( key_holder, param );
    if ( param->suc_lookup ) write_suc_lookup (  key_holder );
    if ( param->unsuc_lookup)write_unsuc_lookup ( key_holder, param );
    if ( param->del )        write_del ( key_holder );
  } /* end of for loop */

} /* end of write_outfile */

/*---main-------------------------------------------------*/
void main( int argc,        /* Number of strings in array argv          */
          char *argv[] )    /* Array of command-line argument strings   */
{

  parameters *param;
    
  table_init();
  param = (Pparameters)malloc(sizeof(parameters));
  param_init( param );

  read_parameters( param );
  write_outfile( param );

  free ( param );
} /* end of main */

